/*
 * sock-inet.c - inet sockets
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <sys/fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <linux/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <errno.h>

#include "xmalloc.h"
#include "log.h"
#include "io.h"
#include "prioq.h"
#include "timeout.h"
#include "sock-unix.h"

#define LISTEN_BACKLOG	100

/* this small struct is used to preserve some data */
struct sock_status {
	int			error_code;
	int			saved_errno;
	struct io_context 	*io;
};

static void connect_event(int fd, short event, void *opaque)
{
	int error;
	socklen_t slen;
	struct sock_status *ss = (struct sock_status *) opaque;

	if (event & EVENT_ERROR) {
		if (!getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &slen)) {
			if (error) {
				debug("sock_inet_connect_wait: connect() failed: %s",
						strerror(ss->error_code));
				ss->saved_errno = error;
				ss->error_code = -1;
			} else {
				warn("sock_inet_connect_wait: connect() failed but "
						"getsockopt() says nothing");
				ss->saved_errno = ECONNRESET;
				ss->error_code = -1;
			}
		} else {
			warn("sock_inet_connect_wait: getsockopt() failed: %s",
					strerror(errno));
			ss->saved_errno = errno;
			ss->error_code = -1;
		}
		io_loop_stop(ss->io);
		return;
	}

	if (event  & EVENT_WRITE) {
		slen = sizeof(error);
		if (!getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &slen)) {
			if (error) {
				debug("sock_inet_connect_wait: connect() failed: %s",
						strerror(errno));
				ss->saved_errno = error;
				ss->error_code = -1;
			} else
				ss->error_code = ss->saved_errno = 0;
		} else {
			debug("sock_inet_connect_wait: getsockopt() failed: %s",
					strerror(errno));
			ss->saved_errno = errno;
			ss->error_code = -1;
		}
		io_loop_stop(ss->io);
		return;
	}

	if (event & EVENT_TIMEOUT) {
		ss->saved_errno = ETIMEDOUT;
		ss->error_code = -1;
		io_loop_stop(ss->io);
		return;
	}
}

void sock_inet_close(int fd)
{
	int res;

	do {
		res = close(fd);
	} while (res == -1 && errno == EINTR);
}

int sock_inet_connect_wait(const char *proto, const char *family,
		const char *host, const char *port, int timeout)
{
	int fd = -1;
	int res;
	struct addrinfo hint;
	struct addrinfo *list, *iter;
	struct io_context *io;
	struct timeout *tout;
	struct sock_status ss;

	if (!host || !port || !proto || !family)
		return -1;

	memset(&hint, 0, sizeof(hint));
	hint.ai_flags = 0;
	hint.ai_protocol = 0;
	hint.ai_canonname = NULL;
	hint.ai_next = NULL;
	hint.ai_addr = NULL;
	if (strcmp(family, "ip") == 0 || strcmp(family, "ip4") == 0)
		hint.ai_family = AF_INET;
	else if (strcmp(family, "ip6"))
		hint.ai_family = AF_INET6;
	else
		return -1;
	if (strcmp(proto, "tcp") == 0)
		hint.ai_socktype = SOCK_STREAM;
	else if (strcmp(proto, "udp") == 0)
		hint.ai_socktype = SOCK_DGRAM;
	else
		return -1;
	res = getaddrinfo(host, port, &hint, &list);
	if (res) {
		if (res == EAI_SYSTEM)
			debug("sock_inet_connect_wait: getaddrinfo()=%d: %s",
					res, strerror(errno));
		else
			debug("sock_inet_connect_wait: getaddrinfo()=%d: %s",
					res, gai_strerror(res));
		return -1;
	}
	io = io_new();
	tout = timeout_new(connect_event, &ss);
	ss.io = io;
	ss.error_code = -1;
	ss.saved_errno = ECONNABORTED;
	for (iter = list; iter != NULL; iter = iter->ai_next) {
		fd = socket(iter->ai_family,
			iter->ai_socktype, iter->ai_protocol);
		if (fd >= 0 && iter->ai_socktype == SOCK_DGRAM)
			break;
		if (fd >= 0) {
			sock_nonblock(fd);
			res = connect(fd, iter->ai_addr, iter->ai_addrlen);
			if (res == -1 && errno != EINPROGRESS) {
				debug("sock_inet_connect_wait: connect()=%d: %s",
						res, strerror(errno));
				sock_block(fd);
				close(fd);
				fd = -1;
				continue;
			}
			if (timeout > 0)
				timeout_add(tout, timeout);
			io_add(io, fd, EVENT_WRITE, connect_event, &ss);
			io_loop_run(io);
			if (timeout_active(tout))
				timeout_del(tout);
			io_del(io, fd);
			sock_block(fd);
			if (ss.error_code == 0)
				break;
			close(fd);
			fd = -1;
		} else
			debug("sock_inet_connect_wait: socket()=%d: %s",
					fd, strerror(errno));
	}
	io_free(&io);
	timeout_free(&tout);
	freeaddrinfo(list);
	if (fd < 0) {
		errno = ss.saved_errno;
		return ss.error_code;
	}
	return fd;
}

/* sock_inet_listen - create INET listening udp or tcp socket
 * address: address to listen. If address is NULL then
 *          INADDR_LOOPBACK will be used.
 * port: port number or service name. Can't be NULL.
 * proto: "udp" or "tcp". Can't be NULL.
 * family: "ip4" or "ip6"
 */
int sock_inet_listen(const char *proto, const char *family, const char *port)
{
	int fd = -1;
	int res, option;
	struct addrinfo hint;
	struct addrinfo *list, *iter;

	if (!port || !proto || !family)
		return -1;

	memset(&hint, 0, sizeof(hint));
	hint.ai_flags = AI_PASSIVE | AI_NUMERICHOST;
	hint.ai_protocol = 0;
	hint.ai_canonname = NULL;
	hint.ai_next = NULL;
	hint.ai_addr = NULL;
	if (strcmp(family, "ip") == 0 || strcmp(family, "ip4") == 0)
		hint.ai_family = AF_INET;
	else if (strcmp(family, "ip6"))
		hint.ai_family = AF_INET6;
	else
		return -1;
	if (strcmp(proto, "tcp") == 0)
		hint.ai_socktype = SOCK_STREAM;
	else if (strcmp(proto, "udp") == 0)
		hint.ai_socktype = SOCK_DGRAM;
	else
		return -1;
	res = getaddrinfo(NULL, port, &hint, &list);
	if (res) {
		if (res == EAI_SYSTEM)
			debug("sock_inet_listen: getaddrinfo()=%d: %s",
					res, strerror(errno));
		else
			debug("sock_inet_listen: getaddrinfo()=%d: %s",
					res, gai_strerror(res));
		return -1;
	}
	for (iter = list; iter != NULL; iter = iter->ai_next) {
		fd = socket(iter->ai_family,
			iter->ai_socktype, iter->ai_protocol);
		if (fd >= 0) {
			option = 1;
			if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
					&option, sizeof(option)))
				warn("sock_inet_listen: setsockopt()=%d: %s",
						fd, strerror(errno));
			res = bind(fd, iter->ai_addr, iter->ai_addrlen);
			if (res == 0) {
				if (iter->ai_socktype == SOCK_DGRAM)
					break;
				res = listen(fd, LISTEN_BACKLOG);
				if (res == 0)
					break;
				debug("sock_inet_listen: listen()=%d: %s",
						fd, strerror(errno));
				close(fd);
				fd = -1;
			} else {
				debug("sock_inet_listen: bind()=%d: %s",
					res, strerror(errno));
				close(fd);
				fd = -1;
			}
		} else {
			debug("sock_inet_listen: socket()=%d: %s",
					fd, strerror(errno));
		}
	}
	freeaddrinfo(list);
	return fd;
}
